#ifndef VECTOR_H_
#define VECTOR_H_

#include <cmath>
#include <string>

namespace arda 
   {

   ////////////////////////////////////////////////////////////////////////////////
   // Vectors
   //
   // Supported operations on vectors
   //
   // assign*()		Assign the value of one vector to another.
   // add*()		Add two vectors
   // cross()		Cross product of two 3-space vectors.
   // dot*()		Dot product.
   // get_angle*()	Calculate angle between two vectors.
   // length*()		Length of a vector (the Euclidean norm).
   // normalize*()	Normalize the vector.
   // proj*()		Calculate the projection of one vector onto another.
   // scale*()		Multiply vector by a scalar.
   // subtract*()	A convenience function to avoid writing add(v1, scale(v2, -1))
   // vectostr*()	Returns a printable string representation of the vector.
   //
   // All functions except vectostr are defined inline.

   namespace Vector 
      {

      typedef int Vector2i[2];
      typedef float Vector2f[2];
      typedef double Vector2d[2];

      typedef int Vector3i[3];
      typedef float Vector3f[3];
      typedef double Vector3d[3];

      typedef int Vector4i[4];
      typedef float Vector4f[4];
      typedef double Vector4d[4];

      //////////////////////////////////////////////////////////////////////////
      // vectostr*(Vector v)
      //	Returns a C string representation of the vector.
      //	This is one of the few non-inline Vector functions.

      std::string & vectostr2(Vector2i v, std::string & s);
      std::string & vectostr2(Vector2f v, std::string & s);
      std::string & vectostr2(Vector2d v, std::string & s);
      std::string & vectostr3(Vector3i v, std::string & s);
      std::string & vectostr3(Vector3f v, std::string & s);
      std::string & vectostr3(Vector3d v, std::string & s);
      std::string & vectostr4(Vector4i v, std::string & s);
      std::string & vectostr4(Vector4f v, std::string & s);
      std::string & vectostr4(Vector4d v, std::string & s);
      
      //////////////////////////////////////////////////////////////////////////
      // scale*(Vector v, scalar s)
      //	Multiply a vector by a scalar.
      //	Returns v so that the result may be passed to other functions.

      inline int * scale2(Vector2i v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline int * scale2(Vector2i v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline int * scale2(Vector2i v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }

      inline float * scale2(Vector2f v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline float * scale2(Vector2f v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline float * scale2(Vector2f v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }

      inline double * scale2(Vector2d v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline double * scale2(Vector2d v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline double * scale2(Vector2d v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }


      inline int * scale3(Vector3i v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline int * scale3(Vector3i v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline int * scale3(Vector3i v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
 
      inline float * scale3(Vector3f v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline float * scale3(Vector3f v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline float * scale3(Vector3f v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }

      inline double * scale3(Vector3d v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline double *  scale3(Vector3d v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline double * scale3(Vector3d v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }


      inline int * scale4(Vector4i v, const int s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline int * scale4(Vector4i v, const float s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline int * scale4(Vector4i v, const double s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }

      inline float * scale4(Vector4f v, const int s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline float * scale4(Vector4f v, const float s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline float * scale4(Vector4f v, const double s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }

      inline double * scale4(Vector4d v, const int s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline double * scale4(Vector4d v, const float s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline double * scale4(Vector4d v, const double s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }


      //////////////////////////////////////////////////////////////////////////
      // dot*(Vector v1, Vector v2)
      //	The dot product of two vectors, which must be the same type
      //	Returns a long for int vectors, a double for floating point vectors.

      inline long dot2(Vector2i v1, Vector2i v2)
	 { return (long) v1[0]*v2[0] + (long) v1[1]*v2[1]; }
      inline double dot2(Vector2f v1, Vector2f v2)
	 { return (double) v1[0]*v2[0] + (double) v1[1]*v2[1]; }
      inline double dot2(Vector2d v1, Vector2d v2)
	 { return v1[0]*v2[0] + v1[1]*v2[1]; }

      inline long dot3(Vector3i v1, Vector3i v2)
	 { return (long) v1[0]*v2[0] + (long) v1[1]*v2[1] + (long) v1[2]*v2[2]; }
      inline double dot3(Vector3f v1, Vector3f v2)
	 { return (double) v1[0]*v2[0] + (double) v1[1]*v2[1] + (double) v1[2]*v2[2]; }
      inline double dot3(Vector3d v1, Vector3d v2)
	 { return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2]; }

      inline long dot4(Vector4i v1, Vector4i v2)
	 { return (long) v1[0]*v2[0] + (long) v1[1]*v2[1] + (long) v1[2]*v2[2] + (long) v1[3]*v2[3]; }
      inline double dot4(Vector4f v1, Vector4f v2)
	 { return (double) v1[0]*v2[0] + (double) v1[1]*v2[1] + (double) v1[2]*v2[2] + (double) v1[3]*v2[3]; }
      inline double dot4(Vector4d v1, Vector4d v2)
	 { return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2] + v1[3]*v2[3]; }


      //////////////////////////////////////////////////////////////////////////
      // length*(Vector v)
      //	Get the length of a vector.
      //	Returns a float for float for float vectors, and a double for 
      //	int and double vectors.

      inline double length2(Vector2i v)
	 { return sqrt((dot2(v, v))); }
      inline double length2(Vector2f v)
	 { return sqrtf((dot2(v, v))); }
      inline double length2(Vector2d v)
	 { return sqrt((dot2(v, v))); }

      inline double length3(Vector3i v)
	 { return sqrt((dot3(v, v))); }
      inline double length3(Vector3f v)
	 { return sqrtf((dot3(v, v))); }
      inline double length3(Vector3d v)
	 { return sqrt((dot3(v, v))); }

      inline double length4(Vector4i v)
	 { return sqrt((dot4(v, v))); }
      inline double length4(Vector4f v)
	 { return sqrtf((dot4(v, v))); }
      inline double length4(Vector4d v)
	 { return sqrt((dot4(v, v))); }


      //////////////////////////////////////////////////////////////////////////
      // add*(Vector v1, Vector v2, Vector vres)
      //	Add two vectors together.
      //        Returns vres so that the result may be used in further 
      //	calculations.

      inline int * add2(Vector2i v1, Vector2i v2, Vector2i vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; return vres; }
      inline float * add2(Vector2f v1, Vector2f v2, Vector2f vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; return vres; }
      inline double * add2(Vector2d v1, Vector2d v2, Vector2d vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; return vres; }

      inline int * add3(Vector3i v1, Vector3i v2, Vector3i vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2]; return vres; }
      inline float * add3(Vector3f v1, Vector3f v2, Vector3f vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2];  return vres; }
      inline double * add3(Vector3d v1, Vector3d v2, Vector3d vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2];  return vres; }

      inline int * add4(Vector4i v1, Vector4i v2, Vector4i vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2]; vres[3] = v1[3] + v2[3]; return vres;}
      inline float * add4(Vector4f v1, Vector4f v2, Vector4f vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2]; vres[3] = v1[3] + v2[3]; return vres; }
      inline double * add4(Vector4d v1, Vector4d v2, Vector4d vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2]; vres[3] = v1[3] + v2[3]; return vres; }

      //////////////////////////////////////////////////////////////////////////
      // normalize*(Vector v)
      //	Normalizes the vecor.  Only defined for floating point vectors.
      //	Returns the vector.

      inline float * normalize2(Vector2f v)
	 { double l = length2(v); v[0] /= l; v[1] /= l; return v; }
      inline double * normalize2(Vector2d v)
	 { double l = length2(v); v[0] /= l; v[1] /= l; return v; }

      inline float * normalize3(Vector3f v)
	 { double l = length3(v); v[0] /= l; v[1] /= l; v[2] /= l; return v; }
      inline double * normalize3(Vector3d v)
	 { double l = length3(v); v[0] /= l; v[1] /= l; v[2] /= l; return v; }

      inline float * normalize4(Vector4f v)
	 { double l = length4(v); v[0] /= l; v[1] /= l; v[2] /= l; v[3] /= l; return v; }
      inline double * normalize4(Vector4d v)
	 { double l = length4(v); v[0] /= l; v[1] /= l; v[2] /= l; v[3] /= l; return v; }

      //////////////////////////////////////////////////////////////////////////
      // cross(Vector3 v)
      //	Only defined for floating point 3 vectors.
      //	Returns the result vector.

      inline float * cross(Vector3f v1, Vector3f v2, Vector3f vres)
	 { 
	 vres[0] =  v1[1]*v2[2] - v2[1]*v1[2];
	 vres[1] = -v1[0]*v2[2] + v2[0]*v1[2];
	 vres[2] =  v1[0]*v2[1] - v2[0]*v1[1];
	 return vres;
	 }

      inline double * cross(Vector3d v1, Vector3d v2, Vector3d vres)
	 { 
	 vres[0] =  v1[1]*v2[2] - v2[1]*v1[2];
	 vres[1] = -v1[0]*v2[2] + v2[0]*v1[2];
	 vres[2] =  v1[0]*v2[1] - v2[0]*v1[1];
	 return vres;
	 }
   
      //////////////////////////////////////////////////////////////////////////
      // get_angle*(Vector3 v1, Vector3 v2)
      //	Returns the angle in radians between the two vectors.
      //
      // get_angle[234]n() take pre-normalized vectors as arguments and will
      //	fail otherwise.  This version is not defined for interger
      //	vectors since you can't normalize an integer vector.
      // get_angle[234]() take arbitrary vectors as arguments.
      //
      // I started out with two version of get_angle3:
      // get_angle3_1 implemented as
      //  return acos(dot3(v1, v2)/(length3(v1)*length3(v2)));
      // get_angle3_2 implemented as
      //  double tmp = dot3(v1, v2); 
      //  return acos(sqrtf(tmp*tmp/(dot3(v1,v1)*dot3(v2,v2))));
      // 
      // I ran timing tests, since I expected the second one - the one I chose to 
      // use, in the end - to be faster since it only resulted in one sqrt call
      // instead of two.  Here are the results
      //
      // get_angle3_1 44.753006890 seconds for 500,000,000 iterations.
      // get_angle3_2 36.940170102 seconds for 500,000,000 iterations.
      //
      // Faster?  Yes. About 7 nanoseconds vs 9.  A ridiculous optimization
      // to worry about?  Eh.  I might as well use it.

      inline double get_angle2n(Vector2f v1, Vector2f v2)
	 { return acos(dot2(v1, v2)); }
      inline double get_angle2n(Vector2d v1, Vector2d v2)
	 { return acos(dot2(v1, v2)); }
      inline double get_angle2(Vector2i v1, Vector2i v2)
	 { 
	 double tmp = dot2(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot2(v1,v1)*dot2(v2,v2)))); 
	 }
      inline double get_angle2(Vector2f v1, Vector2f v2)
	 { 
	 double tmp = dot2(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot2(v1,v1)*dot2(v2,v2)))); 
	 }
      inline double get_angle2(Vector2d v1, Vector2d v2)
	 { 
	 double tmp = dot2(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot2(v1,v1)*dot2(v2,v2)))); 
	 }

      inline double get_angle3n(Vector3f v1, Vector3f v2)
	 { return acos(dot3(v1, v2)); }
      inline double get_angle3n(Vector3d v1, Vector3d v2)
	 { return acos(dot3(v1, v2)); }
      inline double get_angle3(Vector3i v1, Vector3i v2)
	 { 
	 double tmp = dot3(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot3(v1,v1)*dot3(v2,v2)))); 
	 }
      inline double get_angle3(Vector3f v1, Vector3f v2)
	 { 
	 double tmp = dot3(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot3(v1,v1)*dot3(v2,v2)))); 
	 }
      inline double get_angle3(Vector3d v1, Vector3d v2)
	 { 
	 double tmp = dot3(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot3(v1,v1)*dot3(v2,v2)))); 
	 }

      inline double get_angle4n(Vector4f v1, Vector4f v2)
	 { return acos(dot4(v1, v2)); }
      inline double get_angle4n(Vector4d v1, Vector4d v2)
	 { return acos(dot4(v1, v2)); }
      inline double get_angle4(Vector4i v1, Vector4i v2)
	 { 
	 double tmp = dot4(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot4(v1,v1)*dot4(v2,v2)))); 
	 }
      inline double get_angle4(Vector4f v1, Vector4f v2)
	 { 
	 double tmp = dot4(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot4(v1,v1)*dot4(v2,v2)))); 
	 }
      inline double get_angle4(Vector4d v1, Vector4d v2)
	 { 
	 double tmp = dot4(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot4(v1,v1)*dot4(v2,v2)))); 
	 }

      //////////////////////////////////////////////////////////////////////////
      // subtract*(Vector v1, Vector v2, Vector vres)
      //	Subtract v2 from v1 and store result in vres.
      //	Equivalent to add(v1, scale(v2, -1), vres).
      //        Returns vres so that the result may be used in further 
      //	calculations.

      inline int * subtract2(Vector2i v1, Vector2i v2, Vector2i vres)
	 { vres[0] = v1[0] - v2[0]; vres[1] = v1[1] - v2[1]; return vres; }
      inline float * subtract2(Vector2f v1, Vector2f v2, Vector2f vres)
	 { vres[0] = v1[0] - v2[0]; vres[1] = v1[1] - v2[1]; return vres; }
      inline double * subtract2(Vector2d v1, Vector2d v2, Vector2d vres)
	 { vres[0] = v1[0] - v2[0]; vres[1] = v1[1] - v2[1]; return vres; }

      inline int * subtract3(Vector3i v1, Vector3i v2, Vector3i vres)
	 { vres[0] = v1[0] - v2[0]; vres[1] = v1[1] - v2[1]; vres[2] = v1[2] - v2[2]; return vres; }
      inline float * subtract3(Vector3f v1, Vector3f v2, Vector3f vres)
	 { vres[0] = v1[0] - v2[0]; vres[1] = v1[1] - v2[1]; vres[2] = v1[2] - v2[2];  return vres; }
      inline double * subtract3(Vector3d v1, Vector3d v2, Vector3d vres)
	 { vres[0] = v1[0] - v2[0]; vres[1] = v1[1] - v2[1]; vres[2] = v1[2] - v2[2];  return vres; }

      inline int * subtract4(Vector4i v1, Vector4i v2, Vector4i vres)
	 { vres[0] = v1[0] - v2[0]; vres[1] = v1[1] - v2[1]; vres[2] = v1[2] - v2[2]; vres[3] = v1[3] - v2[3]; return vres;}
      inline float * subtract4(Vector4f v1, Vector4f v2, Vector4f vres)
	 { vres[0] = v1[0] - v2[0]; vres[1] = v1[1] - v2[1]; vres[2] = v1[2] - v2[2]; vres[3] = v1[3] - v2[3]; return vres; }
      inline double * subtract4(Vector4d v1, Vector4d v2, Vector4d vres)
	 { vres[0] = v1[0] - v2[0]; vres[1] = v1[1] - v2[1]; vres[2] = v1[2] - v2[2]; vres[3] = v1[3] - v2[3]; return vres; }


      //////////////////////////////////////////////////////////////////////////
      // assign*(Vector v1, Vector v2)
      //	Copies the value of v2 into v1.
      //	Returns v1.

      inline int * assign2(Vector2i v1, Vector2i v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; return v1; }
      inline float * assign2(Vector2f v1, Vector2f v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; return v1; }
      inline double * assign2(Vector2d v1, Vector2d v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; return v1; }

      inline int * assign3(Vector3i v1, Vector3i v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; return v1; }
      inline float * assign3(Vector3f v1, Vector3f v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; return v1; }
      inline double * assign3(Vector3d v1, Vector3d v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; return v1; }

      inline int * assign4(Vector4i v1, Vector4i v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; v1[3] = v2[3]; return v1; }
      inline float * assign4(Vector4f v1, Vector4f v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; v1[3] = v2[3]; return v1; }
      inline double * assign4(Vector4d v1, Vector4d v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; v1[3] = v2[3]; return v1; }


      //////////////////////////////////////////////////////////////////////////
      // proj*(Vector v1, Vector v2, Vector vres)
      //	Calculates the projection of v1 onto v2 and returns the result 
      //	in vres.
      //	Only defined for floating point vectors.
      //	Returns vres.

      inline float * proj2(Vector2f v1, Vector2f v2, Vector2f vres)
	 { assign2(vres, v2); scale2(vres, (dot2(v1, vres)/dot2(vres,vres))); return vres; }
      inline double * proj2(Vector2d v1, Vector2d v2, Vector2d vres)
	 { assign2(vres, v2); scale2(vres, (dot2(v1, vres)/dot2(vres,vres))); return vres; }

      inline float * proj3(Vector3f v1, Vector3f v2, Vector3f vres)
	 { assign3(vres, v2); scale3(vres, (dot3(v1, vres)/dot3(vres,vres))); return vres; }
      inline double * proj3(Vector3d v1, Vector3d v2, Vector3d vres)
	 { assign3(vres, v2); scale3(vres, (dot3(v1, vres)/dot3(vres,vres))); return vres; }

      inline float * proj4(Vector4f v1, Vector4f v2, Vector4f vres)
	 { assign4(vres, v2); scale4(vres, (dot4(v1, vres)/dot4(vres,vres))); return vres; }
      inline double * proj4(Vector4d v1, Vector4d v2, Vector4d vres)
	 { assign4(vres, v2); scale4(vres, (dot4(v1, vres)/dot4(vres,vres))); return vres; }

      } // Vector

   } // namespace arda

#endif // VECTOR_H_
